var BufferWriter = require('buffer-helper');

var LF = "\n".charCodeAt(0)
  , CR = "\r".charCodeAt(0)
  , SP = " ".charCodeAt(0)
  , CO = ":".charCodeAt(0)
  , H = "H".charCodeAt(0)
  , T = "T".charCodeAt(0)
  , P = "P".charCodeAt(0);

var HttpHeader = function(rawHeader) {
  this.rawHeader = rawHeader;
  this.modified = false;

  if (rawHeader != null) {
    this.length = rawHeader.length;
    this._parse();
  }
}

HttpHeader.prototype._isCRLF = function(buffer, index) {
  return (buffer[index] == CR && buffer[index+1] == LF);
}

HttpHeader.prototype._type = function(buff) {
  var i = 0, finished = false;

  while(i < buff.length - 3 && !finished) {
    if (buff[i] == H && buff[i+1] == T && buff[i+2] == T && buff[i+3] == P) {
      finished = true;
    } else {
      i += 1;
    }
  }

  return finished ? "response" : "request";
}

HttpHeader.prototype.get = function(value, str) {
  return str ? this.header[value].toString() : this.header[value];
}

HttpHeader.prototype.set = function(key, value) {
  this.modified = true;

  var val = typeof(value) == 'number' ? String(value) : value;

  if (!(typeof(val) == 'string' || Buffer.isBuffer(val))) {
    throw {
      name: "Invalid argument",
      msg: "Value should be a string or Buffer.",
      info: "Argument was an " + typeof(val)
    };
  }

  this.length += val.length - this.header[key].length;
  this.header[key] = typeof(val) == 'string' ? new Buffer(val) : val;
}

HttpHeader.prototype.template = {
  response: {
    startLine: [
      'version', 'code', 'reason'
    ],
    field: true
  },
  request: {
    startLine: [
      'method', 'uri', 'version'
    ],
    field: true
  }
}

HttpHeader.prototype.getTemplate = function() {
  return this.template[this.header.type];
}

HttpHeader.prototype.toBuffer = function() {
  if (!this.modified) {
    return this.rawHeader;
  }

  var w = new BufferWriter(new Buffer(this.length));
  var t = this.getTemplate();

  w.append(this.header[t.startLine[0]]);
  for (var i = 1; i < t.startLine.length; i++) {
    w.append(SP);
    w.append(this.header[t.startLine[i]]);
  }
  w.appendCRLF();

  for (i = 0; i < this.header.fields.length; i++) {
    w.append([this.header.fields[i].key, CO, this.header.fields[i].value]);
    w.appendCRLF();
  }
  w.appendCRLF();

  w.append(this.body);

  return w.buffer;
}

HttpHeader.prototype._parse = function() {
  this.header = {};
  this.body;

  var tmp;
  tmp = this.parseStartLine(this.rawHeader);
  this.header = tmp.startLine;
  tmp = this.parseHeaders(this.rawHeader, tmp.i + 1);
  this.header.fields = tmp.headers;

  this.body = this.parseBody(this.rawHeader, tmp.i + 1);

  return this.headers;
}

HttpHeader.prototype.startLineVars = {
  request: ["method", "uri", "version"],
  response: ["version", "code", "reason"]
};
HttpHeader.prototype.parseStartLine = function(buff, index) {
  var i = index != null ? index : 0;
  var res = {};
  var token = [], tokens = [], finished = false;

  while (i < buff.length && !finished) {
    var finished = this._isCRLF(buff, i);
    if (buff[i] != SP && !finished) {
      token.push(buff[i]);
    } else {
      tokens.push(new Buffer(token));
      token = [];
    }
    i += 1;
  }

  res.type = this._type(tokens[0]);

  for (var j=0; j < tokens.length; j++) {
    res[this.startLineVars[res.type][j]] = tokens[j];
  }

  return {startLine: res, i: i};
}

HttpHeader.prototype.parseHeaders = function(buff, index) {
  var i = index != null ? index : 0;
  var headers = [];

  var finished = false;
  while(i < buff.length && !finished) {
    if (this._isCRLF(buff, i)) {
      finished = true;
    } else {
      var h = this.parseHeader(buff, i);
      headers.push({
	key: h.key,
	value: h.value
      });
      i = h.i;
    }
    i += 1;
  }

  return {headers: headers, i: i};
}

HttpHeader.prototype.parseHeader = function(buff, index) {
  var i = index != null ? index : 0;
  var key = [], value = [];
  var state = 0;

  while (i < buff.length && state < 3) {
    if (state == 0) {
      if (buff[i] == CO) {
	state = 1;
      } else {
	key.push(buff[i]);
      }
    } else if (state == 1) {
      if (this._isCRLF(buff, i)) {
	state = 3;
      } else {
	value.push(buff[i]);
      }
    }
    i += 1;
  }

  var res = {
    key: new Buffer(key),
    value: new Buffer(value),
    i: i
  };
  return res;
}

HttpHeader.prototype.parseBody = function(buff, index) {
  return buff.slice(index, buff.length);
}

module.exports = {
  HttpHeader: HttpHeader
}
